// ************************************************************
// *****             CPTS 562 - Final Project             *****
// *****             4-Bit RC4 Cipher Breaker             *****
// *****         05/02, 05/03/2002 by Stefan Witt         *****
// ************************************************************

import java.io.*;



/**
  *  This class implements a cipher breaker for the 4-bit version of RC4.
  *  It performs a brute force attack to try all of the 15! possible
  *  key permutations.<BR>
  *  Note that RC4 is a patent of RSA Data Security Inc.
  *  @author  Stefan Witt
  *  @version 1.0
  *  @see     RC4
  */

public class RC4Breaker {
    private static final int BITS = 4;           // number of bits (this is also hardcoded below!)
    private static final int BOX  = (1 << BITS); // size of S-Box
    private int              p[];                // current permutation
    private int              plainText[];        // known plain text
    private String           inFile;             // encrypted file
    private String           outFile;            // where to store decrypted output
    private int              cipher[];           // encrypted text


    // ***** Constructor *****
    /**
      *  Create a new RC4Breaker. The breaker does a known plain text attack,
      *  so a plain text has to be passed.
      *  @param knownPlainText a known plain text within the encrypted data
      */

    public RC4Breaker(String knownPlainText) {
	int i;

	plainText = new int[knownPlainText.length()];      // convert plain text to integer array
	for(i = 0; i < knownPlainText.length(); i++) {
	    plainText[i] = (int) knownPlainText.charAt(i);
	}
    }


    // ***** Method run *****
    /**
      *  Perform the RC4 cipher breaking. An encrypted input file is read,
      *  and each possible encryption key is tested if the decrypted file
      *  contains the known plain text. If the text is found, then the
      *  corresponding encrypted data is saved in the output file.
      *  @param inFile  input file
      *  @param outFile where to store the output
      *  @return        true, if the attack was successful, false otherwise
      */

    public boolean run(String inFile, String outFile) {
	File             file;
	FileInputStream  in;
	FileOutputStream out;
	int              length;
	int              i;
	int              a[];

	this.inFile  = inFile;                      // where to get the input
	this.outFile = outFile;                     // where to store the output
	try {
	    file = new File(inFile);                // examine file to determine its length
	    length = (int) file.length();           // get file size
	    cipher = new int[length];               // create byte array of file size
	    in = new FileInputStream(inFile);       // open file as stream
	    for (i = 0; i < length; i++) {
		cipher[i] = in.read();              // read file to array
	    }
	    in.close();                             // close file again
	} catch (Exception e) {
	    System.err.println("RC4Breaker.run() I/O error; " + e.getMessage());
	    return(false);
	}
	p = new int[BOX];                           // initialize permutation array
	a = new int[BOX];                           // array to be permuted
	for (i = 0; i < BOX; i++) {
	    a[i] = i;
	}
	return(permute(a, 0));                      // run cipher breaker
    }


    // ***** Method permute (private) *****
    /**
      *  This method recursively creates all possible permutations of the input array.
      *  Whenever a permutation is generated, the method processPermutation() is called.
      *  @param restPerm the array to be permuted
      *  @param depth    recursion depth (start with 0)
      *  @return         true, if cipher broken, false otherwise
      */

    private boolean permute(int restPerm[], int depth) {
	int i;         // counters
	int j;
	int k;
	int subPerm[]; // for creating array for recursive call

	if (restPerm.length == 1) {                     // trivial case: only one choice
	    p[depth] = restPerm[0];                     // apply the choice
	    return(processPermutation(p));              // permutation complete, process it
	} else {                                        // non-trivial case: use recursion
	    subPerm = new int[restPerm.length - 1];     // set up array for recursive call
	    for (i = 0; i < restPerm.length; i++) {     // pick each element of restPerm[] once
		j = 0;
		for (k = 0; k < restPerm.length; k++) { // copy restPerm[] to subPerm[]
		    if (k == i) {                       // but skip picked element
			continue;
		    }
		    subPerm[j] = restPerm[k];
		    j++;
		}
		p[depth] = restPerm[i];                 // copy picked element to output permutation
		if (permute(subPerm, depth + 1)) {      // do recursive call
		    return(true);                       // done if success
		}
	    }
	}
	return(false);                                  // not done yet
    }


    // ***** Method processPermutation (private) *****
    /**
      *  This method tries to do a known plain text on encrypted data.
      *  With a given key it decrypts the data and looks for the plain text.
      *  If the plain text is found, the complete data is decrypted and saved
      *  to an output file.
      *  @param permutation the permutation to try the decryption with
      *  @return            true, if cipher broken, false otherwise
      */

    private boolean processPermutation(int permutation[]) {
	RC4 r;
	int cipherOffset;
	int plainOffset;
	int decrypt;
	int i;

	System.out.print("Trying key: ");                        // print used key
	for (i = 0; i < BOX; i++) {
	    System.out.print(permutation[i] + " ");
	}
	System.out.println("");
	r = new RC4(permutation);                                 // create RC4 random number generator with the permutation
	plainOffset = 0;
	for (cipherOffset = 0; cipherOffset < cipher.length; cipherOffset++) {
	    decrypt = cipher[cipherOffset] ^ r.getRandomNumber(); // decrypt one byte
	    if (decrypt == plainText[plainOffset]) {              // check if it matches the plain text
		plainOffset++;                                    // yes, one char is found
		if (plainOffset >= plainText.length) {            // complete plain text found?
		    r = new RC4(permutation);                     // yes, decrypt the whole file
		    r.run(inFile, outFile);
		    return(true);                                 // success
		}
	    } else {
		plainOffset = 0;                                  // char does not match, try again
		if (decrypt == plainText[plainOffset]) {          // check again for first character
		    plainOffset++;                                // yes, one char is found
		    if (plainOffset >= plainText.length) {        // complete plain text found?
			r = new RC4(permutation);                 // yes, decrypt the whole file
			r.run(inFile, outFile);
			return(true);                             // success
		    }
		}
	    }
	}
	return(false);
    }


    // ***** Method main *****
    /**
      *  This program runs the RC4 cipher breaker on an encrypted file and
      *  tries every possible RC4 key on it to search for a known plain text.
      *  If the cipher breaking is successful, then the decrypted data is saved in
      *  a specified output file.<BR>
      *  Invoke with "java RC4 inputfile outputfile". If no 2 files are
      *  specified, they are prompted.
      *  @param argv command line arguments
      */

    public static void main(String argv[]) {
	String         beforeFile;
	String         afterFile;
	String         knownText;
	BufferedReader stdin;
	RC4Breaker     r;

	try {
	    stdin = new BufferedReader(new InputStreamReader(System.in));
	    if (argv.length != 2) {
		System.out.print("Input file: ");         // ask for files if not specified
		beforeFile = stdin.readLine();
		System.out.print("Output file: ");
		afterFile = stdin.readLine();
	    } else {
		beforeFile = argv[0];                     // otherwise use given names
		afterFile = argv[1];
	    }
	    System.out.print("Enter known plain text: "); // prompt for some plain text for attack
	    knownText = stdin.readLine();
	    r = new RC4Breaker(knownText);                // run cipher breaker
	    if (r.run(beforeFile, afterFile)) {
		System.out.println("RC4 breaking successful");
	    } else {
		System.out.println("RC4 breaking unsuccessful");
	    }
	} catch (IOException e) {
	    System.err.println("couldn't read from stdin");
	    System.exit(1);
	}
    }
}
