// ************************************************************
// *****             CPTS 562 - Final Project             *****
// *****        4-Bit RC4-Encryption and Decryption       *****
// *****         05/01, 05/02/2002 by Stefan Witt         *****
// ************************************************************

import java.io.*;



/**
  *  This class implements a 4-bit version of RC4. RC4 is a very fast random
  *  number generator, which can be used for stream cipher, like in this
  *  program.<BR>
  *  Note that RC4 is a patent of RSA Data Security Inc.
  *  @author  Stefan Witt
  *  @version 1.0
  */

public class RC4 {
    private static final int BITS = 4;           // number of bits (this is also hardcoded below!)
    private static final int BOX  = (1 << BITS); // size of S-Box
    private int    K[];                          // cipher key
    private int    S[];                          // S-Box
    private int    i;                            // state i of RC4
    private int    j;                            // state j of RC4


    // ***** Constructor *****
    /**
      *  This constructs a 4-Bit RC4 random number generator. To initialize
      *  the start value, a secret string "salt" must be passed.
      *  @param salt secret string to generate start value for RC4
      */

    public RC4(String salt) {
	this.K = new int[BOX];                  // set up key
	generateKey(salt);                      // key perturbation
	init();
    }


    // ***** Constructor *****
    /**
      *  This constructs a 4-Bit RC4 random number generator. To initialize
      *  the start value, a secret key (an integer array of a permutation of
      *  0,..., 15) must be passed.
      *  @param key integer array of size 16 with a permutation of 0,..., 15
      */

    public RC4(int key[]) {
	int count; // counter ("i" and "j" are already the RC4 states)

	this.K = new int[BOX];                  // set up key
	for (count = 0; count < BOX; count++) { // copy key
	    K[count] = key[count];
	}
	init();                                 // create start S-Box
    }


    // ***** Method init (private) *****
    /**
      *  This initializes the S-Box of the RC4 generator using the RC4 algorithm.
      */

    private void init() {
	int count; // counter ("i" and "j" are already the RC4 states)
	int swap;

	this.S = new int[BOX];                  // set up S-Box
	for (count = 0; count < BOX; count++) {
	    S[count] = count;
	}
	j = 0;                                  // initialize S-Box
	for (i = 0; i < BOX; i++) {
	    j = (j + S[i] + K[i]) % BOX;
	    swap = S[i];
	    S[i] = S[j];
	    S[j] = swap;
	}
	i = 0;
	j = 0;
    }


    // ***** Method generateKey (private) *****
    /**
      *  This generates the initial key for initializing the S-Box.
      *  @param salt secret string to generate initial key
      */

    private void generateKey(String salt) {
	int count;
	int swap;
	int a;
	int b;

	for (count = 0; count < BOX; count++) {                  // initialize
	    K[count] = count;
	}
	for (count = 0; count < salt.length() - 1; count += 2) { // perturb according to string
	    a = ((int) salt.charAt(count)) & (BOX - 1);          // use lower 4 bits
	    b = ((int) salt.charAt(count + 1)) & (BOX - 1);
	    swap = K[a];
	    K[a] = K[b];
	    K[b] = swap;
	}
    }


    // ***** Method getRandomNumber *****
    /**
      *  Generate an random number between 0 and 255.
      *  @return the random number
      */

    public int getRandomNumber() {
	int t;
	int number;
	int swap;

	i = (i + 1) % BOX;       // generate upper 4 bits (with RC4 algorithm)
	j = (j + S[i]) % BOX;
	swap = S[i];
	S[i] = S[j];
	S[j] = swap;
	t = (S[i] + S[j]) % BOX;
	number = S[t] << BITS;
	i = (i + 1) % BOX;       // generate lower 4 bits
	j = (j + S[i]) % BOX;
	swap = S[i];
	S[i] = S[j];
	S[j] = swap;
	t = (S[i] + S[j]) % BOX;
	number |= S[t];
	return(number);          // and return number
    }


    // ***** Method run *****
    /**
      *  Run RC4 stream cipher on a file. Since this is symmetric,
      *  it can be used to encrypt a file and to decrypt an encrypted file.
      *  @param inFile  the input file
      *  @param outFile the output file
      */

    public void run(String inFile, String outputFile) {
	FileInputStream  in;
	FileOutputStream out;
	int              b;

	System.out.print("Key: ");                  // print used key
	for (b = 0; b < BOX; b++) {
	    System.out.print(K[b] + " ");
	}
	System.out.println("");
	try {
	    in = new FileInputStream(inFile);       // open both files
	    out = new FileOutputStream(outputFile);
	    b = in.read();                          // read first byte
	    while (b != -1) {                       // until end of file
		b ^= getRandomNumber();             // encrypt byte with stream cipher
		out.write(b);                       // write it to output
		b = in.read();                      // next byte
	    }
	    out.close();                            // that's it
	    in.close();
	} catch (Exception e) {
	    System.err.println("RC4.run() I/O error; " + e.getMessage());
	}
    }


    // ***** Method main *****
    /**
      *  Run RC4 stream cipher on a file and write result to output file.
      *  Since this is symmetric, it can be used to encrypt a file and
      *  to decrypt an encrypted file.<BR>
      *  Invoke with "java RC4 inputfile outputfile". If no 2 files are
      *  specified, they are prompted.
      *  @param argv command line arguments
      */

    public static void main(String argv[]) {
	String         beforeFile;
	String         afterFile;
	String         salt;	BufferedReader stdin;
	RC4            r;

	try {
	    stdin = new BufferedReader(new InputStreamReader(System.in));
	    if (argv.length != 2) {
		System.out.print("Input file: ");    // ask for files if not specified
		beforeFile = stdin.readLine();
		System.out.print("Output file: ");
		afterFile = stdin.readLine();
	    } else {
		beforeFile = argv[0];                // otherwise use given names
		afterFile = argv[1];
	    }
	    System.out.print("Enter pass phrase: ");
	    salt = stdin.readLine();
	    r = new RC4(salt);                       // run RC4 on them
	    r.run(beforeFile, afterFile);
	} catch (IOException e) {
	    System.err.println("couldn't read from stdin");
	    System.exit(1);
	}
    }
}
