// ************************************************************
// *****             CPTS 562 - Final Project             *****
// *****   JavaSpaces 4-Bit RC4 Cipher Breaker (Worker)   *****
// *****         05/03, 05/04/2002 by Stefan Witt         *****
// ************************************************************

import java.rmi.*;
import net.jini.space.*;
import net.jini.lease.*;
import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;



/**
  *  This class implements a cipher breaker for the 4-bit version of RC4.
  *  It reads the cipher text from a JavaSpace and reads tasks containing
  *  test keys and known plain texts from the JavaSpace and puts the
  *  results back.
  *  Note that RC4 is a patent of RSA Data Security Inc.
  *  @author  Stefan Witt
  *  @version 1.0
  */

public class RC4BreakerWorker {
    private int       cipher[]; // encrypted text
    private JavaSpace js;       // JavaSpaces service


    // ***** Method doTask *****
    /**
      *  Run a cipher breaking attempt according to a task.
      *  @param task the task to be done
      */

    private boolean doTask(RC4Task task) {
	int i;
	int plainText[];
	int BOX = 16;
	RC4 r;
	int plainOffset;
	int cipherOffset;
	int decrypt;

	System.out.print("computing task (key=");
	for (i = 0; i < BOX; i++) {
	    System.out.print(task.key[i] + " ");
	}
	System.out.print("text=" + task.plainText + ")... ");
	plainText = new int[task.plainText.length()];            // convert plain text to integer array
	for(i = 0; i < task.plainText.length(); i++) {
	    plainText[i] = (int) task.plainText.charAt(i);
	}
	r = new RC4(task.key);                                 // create RC4 random number generator with the permutation
	plainOffset = 0;
	for (cipherOffset = 0; cipherOffset < cipher.length; cipherOffset++) {
	    decrypt = cipher[cipherOffset] ^ r.getRandomNumber(); // decrypt one byte
	    if (decrypt == plainText[plainOffset]) {              // check if it matches the plain text
		plainOffset++;                                    // yes, one char is found
		if (plainOffset >= plainText.length) {            // complete plain text found?
		    System.out.println("true");
		    return(true);                                 // success
		}
	    } else {
		plainOffset = 0;                                  // char does not match, try again
		if (decrypt == plainText[plainOffset]) {          // check again for first character
		    plainOffset++;                                // yes, one char is found
		    if (plainOffset >= plainText.length) {        // complete plain text found?
			System.out.println("true");
			return(true);                             // success
		    }
		}
	    }
	}
	System.out.println("false");
	return(false);
    }


    // ***** Method work *****
    /**
      *  Start worker for RC4 breaking.
      */

    public void work() {
	int                 i;
	RC4Task             tTemplate;
	RC4Task             task;
	TransactionManager  txm;
	LeaseRenewalManager lrm;
	ServiceFinder       sfjs;

	System.setSecurityManager(new RMISecurityManager());                // set up RMI
	try {
	    ServiceFinder sf = new ServiceFinder(TransactionManager.class); // find the transaction manager
	    txm = (TransactionManager)(sf.getObject());
	    lrm = new LeaseRenewalManager();
	    sfjs = new ServiceFinder(JavaSpace.class);                      // find the JavaSpaces
	} catch (Exception e) {
	    System.err.println("RC4BreakerWorker.work() error: " + e.getMessage());
	    e.printStackTrace();
	    return;
	}
	js = (JavaSpace) sfjs.getObject();
	CipherText ctTemplate = new CipherText();                           // obtain cipher text from space
	try {
	    System.out.print("reading ciphertext from JavaSpace... ");
	    CipherText ct = (CipherText) js.read(ctTemplate, null, Long.MAX_VALUE);
	    System.out.println("ok");
	    cipher = new int[ct.cipher.length];                           // and make copy from it
	    for (i = 0; i < ct.cipher.length; i++) {
		cipher[i] = ct.cipher[i];
	    }
	} catch (Exception e) {
	    System.err.println("RC4BreakerWorker.main() error: " + e.getMessage());
	    return;
	}
	while (true) {
	    Transaction.Created txn = null;
	    try {
		// process the task in a transaction so that if this worker
		// exits before it writes the answer the task is not lost
		txn = TransactionFactory.create(txm, Lease.FOREVER);
		lrm.renewUntil(txn.lease, Lease.FOREVER, null);
		System.out.print("taking task from JavaSpace... ");
		tTemplate = new RC4Task();
		tTemplate.done = new Boolean(false);
		task = (RC4Task) js.take(tTemplate, txn.transaction, Long.MAX_VALUE);
		System.out.println("ok");
		task.plainTextFound = new Boolean(doTask(task)); // compute task
		task.done = new Boolean(true);
		System.out.print("writing result to JavaSpace... ");
		js.write(task, txn.transaction, 60 * 60 * 1000);
		System.out.print("ok\ncommiting... ");
		txn.transaction.commit();
		System.out.println("ok");
		lrm.remove(txn.lease);
	    } catch (Exception e) {
		// These exceptions mean that we couldn't talk to the
		// space or transaction manager, which we can't recover from.
		System.out.println("RC4BreakerWorker.work() error: " + e.getMessage());
		e.printStackTrace();
		return;
	    }
	}
    }


    // ***** Method main *****
    /**
      *  This program runs the RC4 cipher breaker on an encrypted file and
      *  tries RC4 keys on it to search for a known plain text.<BR>
      *  It reads the cipher text and cipher breaking tasks from a JavaSpace.
      *  Invoke with "make run".
      *  @param argv command line arguments (ignored)
      */

    public static void main(String[] args) {
	RC4BreakerWorker    r;

	r = new RC4BreakerWorker(); // create worker
	r.work();                   // and let it work
    }
}
